from datetime import datetime
from math import ceil
from .auth import auth
from flask import Blueprint, redirect, render_template, request, url_for

from .db import get_db

bp = Blueprint('jobs', __name__, url_prefix='/jobs')

def parse_date(dt, is_end=False):
  sp = str(dt).split('-')
  if len(sp) >= 3:
    if is_end:
      return datetime(int(sp[0]), int(sp[1]), int(sp[2]), 23, 59, 59, 999)
    else:
      return datetime(int(sp[0]), int(sp[1]), int(sp[2]), 0, 0, 0, 0)

  return None

@bp.route('/', methods=('GET', 'POST'))
@auth.login_required
def items():
  db = get_db()
  jobs=[]
  total=0

  options = dict(
    txt="",
    df="",
    dt="",
    lang="all",
    status="all",
    limit=10,
    page=1,
    total=0,
    numOfPages=1
  )

  if request.method=="POST":
    fd = request.form
    options["txt"] = fd.get("searchText", "")
    options["df"] = fd.get("periodFrom", "")
    options["dt"] = fd.get("periodTo", "")
    options["lang"] = fd.get("language", "all")
    options["status"] = fd.get("status", "all")
    options["limit"] = fd.get("limit", 0)
    options["page"] = fd.get("page", 1)

    if options["page"] in ["<" , "<<", ">", ">>"]:
      cp = int(fd.get("currentPage", 1))
      nps = int(fd.get("numOfPages", 1))
      if options["page"] == "<":
        options["page"] = max(1, cp - 1)
      elif options["page"] == "<<":
        options["page"] = 1
      elif options["page"] == ">":
        options["page"] = min(cp + 1, nps)
      elif options["page"] == ">>":
        options["page"] = nps
    
    options["page"] = int(options["page"])

  query = {}

  if options["df"] != "" or options["dt"] != "":
    query["createdAt"] = {}
    if options["df"] != "":
      d = parse_date(options["df"], False)
      if not d is None:
        query["createdAt"]["$gte"] = d
    if options["dt"] != "":
      d = parse_date(options["dt"], True)
      if not d is None:
        query["createdAt"]["$lte"] = d

  if options["lang"] != "all":
    query["lang"] = options["lang"]

  if options["status"] != "all":
    if options["status"] == "error":
      query["error"] = True
    elif options["status"] == "in-progress":
      query["inProgress"] = True
    elif options["status"] == "done":
      query["done"] = True
    elif options["status"] == "pending":
      query["done"] = {"$ne": True}
      query["error"] = {"$ne": True}
      query["inProgress"] = {"$ne": True}
  
  if options["txt"] != "":
    query["$or"] = [
      {"data.function_name": {"$regex": options["txt"], "$options": "i"}},
      {"lang": {"$regex": options["txt"], "$options": "i"}},
      {"channel": {"$regex": options["txt"], "$options": "i"}}
    ]

  skip = 0
  if options["limit"] != "all":
    skip = (options["page"] - 1) * int(options["limit"])
  
  jobs, total = db.get_jobs(query=query, limit=options["limit"], skip=skip)
  options["total"] = total
  options["numOfPages"] = 1 if options['limit'] == "all" else ceil(int(total) / int(options["limit"]))

  for job in jobs:
    job["job_name"] = job.get("data", {}).get("function_name")
    job["status"] = "pending"

    if job.get("error"):
      job["status"] = "failed"
    elif job.get("done"):
      job["status"] = "completed"
    elif job.get("inProgress"):
      job["status"] = "in progress"

  options["paging"] = {"left": range(max(1, options['page'] - 5), options['page']), "right": range(options['page']+1, min(options["numOfPages"], options['page'] + 5)+1)}
  return render_template('jobs/list.html', jobs=jobs, options=options)


@bp.route('/<jobid>', methods=('GET', 'POST'))
@auth.login_required
def item(jobid):
  db = get_db()
  job = db.get_job(job_id=jobid)
  if job is None:
    return redirect(url_for('jobs.items'))

  job["job_name"] = job.get("data", {}).get("function_name")
  job["status"] = "pending"

  if job.get("error"):
    job["status"] = "failed"
  elif job.get("done"):
    job["status"] = "completed"
  elif job.get("inProgress"):
    job["status"] = "in progress"

  if request.method == 'GET':
    attrs = [
      dict(text="ID", value=job.get("_id")),
      dict(text="Callback", value=job.get("job_name")),
      dict(text="Status", value=job.get("status")),
      dict(text="Channel", value=job.get("channel")),
      dict(text="Language", value=job.get("lang")),
      dict(text="Created At", value=job.get("createdAt")),
      dict(text="Expires At", value=job.get("expireAt")),
    ]

    last_error = job.get("errorMessage")
    errors = []
    
    if job["status"] == "failed" and not last_error is None:
      if not isinstance(last_error, dict):
        last_error = dict(message=last_error, trace="")

      errors = [
        dict(text="Message", value=last_error.get("message"), cssclass="w3-text-red"),
        dict(text="Trace", value=last_error.get("trace"), cssclass="w3-text-red")
      ]

    args = job.get("data", {}).get("args", [])
    k_wargs = job.get("data", {}).get("kwargs", {})
    kwargs = []
    for k in k_wargs:
      kwargs.append(dict(text=k, value=k_wargs.get(k)))

    res = job.get("result")
    has_results = False
    results = []
    if not res is None:
      has_results = True
      if isinstance(res, dict):
        for k in res:
          results.append(dict(text=k, value=res.get(k)))
      elif isinstance(res, list) or isinstance(res, tuple):
        for k, v in enumerate(res):
          results.append(dict(text=k, value=v))
      else:
        results.append(dict(text="", value=res))

    return render_template('jobs/item.html', job=job, attributes=attrs, errors=errors, has_errors=len(errors) > 0, args=args, has_args=len(args) > 0, kwargs=kwargs, has_kwargs=len(kwargs) > 0, has_results=has_results, results=results)


@bp.route('/<jobid>/requeue', methods=['POST',])
@auth.login_required
def requeue(jobid):
  db = get_db()
  job = db.get_job(job_id=jobid)

  if not job is None:
    db.requeue_job(job_id=jobid)
  
  return redirect(url_for('jobs.items'))


@bp.route('/<jobid>/remove', methods=['POST',])
@auth.login_required
def remove(jobid):
  db = get_db()
  job = db.get_job(job_id=jobid)

  if not job is None:
    db.remove_job(job_id=jobid)
  
  return redirect(url_for('jobs.items'))
  


from .app import create_app

def parse_args():
  import argparse
  parser = argparse.ArgumentParser(description='Dashboard for the Mongo Queuing and Scheduling Library', add_help=True)
  parser.add_argument('-u', '--conn', dest='db_conn', type=str, default="mongodb://localhost:27017",  help='mongodb connection string (default: mongodb://localhost:27017)')
  parser.add_argument('--dbname', dest='db_name', type=str, default="jobs",  help='mongodb database name (default: jobs)')
  parser.add_argument('--colname', dest='col_name', type=str, default="jobs",  help='mongodb collection name (default: jobs)')
  parser.add_argument('--auth', dest='auth_file', type=str, default=None,  help='path to a json file which contains list of user accounts for basic authentication')
  parser.add_argument('--debug', dest='debug_mode', action='store_true', help='run flask in debug mode')

  args = parser.parse_args()

  return args

def main():
  import os
  args = parse_args()

  env = {"DB_CONNECTION": args.db_conn, "DB_NAME": args.db_name, "DB_COL_NAME": args.col_name, "AUTH_FILE": args.auth_file, "FLASK_APP": "mqas_dashboard"}
  for k in env:
    if not env[k] is None:
      os.environ[k] = env[k]

  app = create_app()
  app.run(debug=args.debug_mode, use_reloader=False, host="0.0.0.0")



from flask import Flask, redirect, url_for

def create_app(test_config=None):
    # create and configure the app
    app = Flask(__name__, instance_relative_config=True)

    # a simple page that says hello
    @app.route('/health')
    def health():
        return "Application is Running"


    @app.route('/')
    def index():
        return redirect(url_for('jobs.items'))


    from . import workers
    app.register_blueprint(workers.bp)

    from . import jobs
    app.register_blueprint(jobs.bp)

    return app


from flask_httpauth import HTTPBasicAuth
from werkzeug.security import generate_password_hash, check_password_hash
import os, json

auth = HTTPBasicAuth()

def load_users():
  fn = os.environ.get("AUTH_FILE")
  users = {}
  if not fn is None:
    if fn.endswith(".json"):
      with open(fn, "r") as json_file:
        data = json.load(json_file)
        if isinstance(data, dict):
          for k in data:
            users[k] = generate_password_hash(data[k])
  return users

@auth.verify_password
def verify_password(username, password):
  users = load_users()
  if username in users and check_password_hash(users.get(username), password):
    return username


from pymongo import DESCENDING, ASCENDING
from pymongo.collection import Collection
from pymongo.database import Database
from pymongo import MongoClient
from bson import ObjectId
from typing import Union, Optional, Type
from datetime import datetime, timedelta
from flask import g
import os

class DB:
  def __init__(self, connection: Union[str, Type[Collection], Type[Database], Type[MongoClient]] = None, db_name: Optional[str]="jobs", col_name: Optional[str]="jobs") -> None:
    
    if connection is None:
      connection = "mongodb://localhost:27017"

    if isinstance(connection, MongoClient):
      db = connection[db_name]
      self.collection = db[col_name]
    elif isinstance(connection, Database):
      self.collection = connection[col_name]
    elif isinstance(connection, Collection):
      self.collection = connection
    elif isinstance(connection, str):
      self.collection = MongoClient(connection)[db_name][col_name]
    else:
      print("Connection object should be one of either a mongoclient, database or collection objects, or a string representing the connection url", "\n", "Current object type is", type(connection))
      self.collection = connection

  def get_workers(self):
    items = self.collection.find({"is_worker": True}).sort("createdAt", DESCENDING)
    return list(items)

  def get_worker(self, worker_id):
    item = self.collection.find_one({"is_worker": True, "_id": ObjectId(worker_id)})
    return item

  def get_jobs(self, limit=0, skip=0, query={}):
    query["is_worker"] = {"$ne": True}
    
    if limit == "all":
      limit = 0
    else:
      limit = int(limit)

    count = self.collection.count_documents(query)
    items = self.collection.find(query).sort("createdAt", DESCENDING).sort("updatedAt", DESCENDING).skip(skip).limit(limit)
    return list(items), count

  def get_job(self, job_id):
    item = self.collection.find_one({"is_worker": {"$ne": True}, "_id": ObjectId(job_id)})
    return item

  def requeue_job(self, job_id):
    job = self.get_job(job_id)
    expireAt = None
    ttl = job.get("ttl")
    if not ttl is None:
      if int(ttl) > 0:
        expireAt = datetime.utcnow() + timedelta(seconds=int(ttl))

    res = self.collection.update_one({"is_worker": {"$ne": True}, "_id": ObjectId(job_id)}, {"$set": {"done": False, "error": False, "inProgress": False, "errorMessage": None, "result": None, "expireAt": expireAt}})
    return res

  def remove_job(self, job_id):
    res = self.collection.delete_one({"is_worker": {"$ne": True}, "_id": ObjectId(job_id)})
    return res

def get_variables():
  connection = os.environ.get("DB_CONNECTION", "mongodb://localhost:27017")
  db_name = os.environ.get("DB_NAME", "jobs")
  col_name = os.environ.get("DB_COL_NAME", "jobs")
  return dict(connection=connection, db_name=db_name, col_name=col_name)

def get_db():
  if 'db' not in g:
    kwargs = get_variables()
    g.db = DB(**kwargs)

  return g.db


from .auth import auth
from flask import Blueprint, redirect, render_template, request, url_for
from .db import get_db

bp = Blueprint('workers', __name__, url_prefix='/workers')

@bp.route('/', methods=('GET', 'POST'))
@auth.login_required
def items():
  db = get_db()
  workers = db.get_workers()
  if request.method == 'GET':
    return render_template('workers/list.html', workers=workers)


@bp.route('/<workerid>', methods=('GET', 'POST'))
@auth.login_required
def item(workerid):
  db = get_db()
  worker = db.get_worker(worker_id=workerid)
  if worker is None:
    return redirect(url_for('workers.items'))

  if request.method == 'GET':
    attrs = [
      dict(text="ID", value=worker.get("_id")),
      dict(text="System Info", value=worker.get("info")),
      dict(text="Status", value=worker.get("status")),
      dict(text="Queues", value=worker.get("queues")),
      dict(text="Completed Jobs", value=worker.get("jobs_completed")),
      dict(text="Failed Jobs", value=worker.get("jobs_failed")),
      dict(text="Created At", value=worker.get("createdAt")),
      dict(text="Updated At", value=worker.get("updatedAt")),
    ]

    last_error = worker.get("last_error_message")
    errors = []
    if not last_error is None:
      if not isinstance(last_error, dict):
        last_error = dict(message=last_error, trace="")

      errors = [
        dict(text="Job Id", value=last_error.get("jobId"), cssclass="w3-text-red"),
        dict(text="Message", value=last_error.get("message"), cssclass="w3-text-red"),
        dict(text="Occurred At", value=last_error.get("errorAt"), cssclass="w3-text-red"),
        dict(text="Trace", value=last_error.get("trace"), cssclass="w3-text-red"),
        dict(text="Job", value=last_error.get("callback"), cssclass="w3-text-red"),
      ]

    return render_template('workers/item.html', worker=worker, attributes=attrs, errors=errors, has_errors=len(errors) > 0)
